#include "main.h"
#include <stdio.h>
#include <string.h>

void consTask(struct sheet *sht)
{
	struct task *task = taskNow();
	struct console cons;
	struct fhandle fh[8];
	int i;
	for (i = 0; i < 8; i++) {
		fh[i].buf = 0;
	}
	cons.sht = sht;
	cons.cx = 8;
	cons.cy = 28;
	cons.cc = -1;
	task->cons = &cons;
	task->fh = fh;
	consPutc(&cons, '>', 1);

	for (;;) {
		cli();
		if (fifoStat(&task->fifo) == 0) {
			taskSleep(task);
			sti();
		} else {
			i = fifoGet(&task->fifo);
			sti();
			if (i == 2) {	/* J[\ON */
				cons.cc = _FFFFFF;
			} else if (i == 3) {	/* J[\OFF */
				if (cons.sht != 0) {
					consPutc(&cons, ' ', 0);
				}
				cons.cc = -1;
			} else if (i == 4) {
				consExit(&cons, task->fat);
			} else if (256 <= i && i <= 511) {
				i -= 256;
				if (i == 8) {
					if (cons.cx > 16) {
						consPutc(&cons, ' ', 0);
						cons.cx -= 8;
					}
				} else if (i == 10) {
					cons.line[cons.cx / 8 - 2] = 0;
					consPutc(&cons, ' ', 0);
					consNline(&cons);
					consRuncmd(&cons, task->fat);
					consNline(&cons);
					consPutc(&cons, '>', 1);
					if (cons.sht == 0) {
						consExit(&cons, task->fat);
					}
				} else {
					if (cons.cx < 240) {
						cons.line[cons.cx / 8 - 2] = i;
						consPutc(&cons, (unsigned char) i, 1);
					}
				}
			}
			if (cons.sht != 0) {
				if (cons.cc >= 0) {
					graphicBox(sht, cons.cc, cons.cx, cons.cy, cons.cx + 7, cons.cy + 15);
				}
				shtRefresh(sht, cons.cx, cons.cy, cons.cx + 8, cons.cy + 16);
			}
		}
	}
}

struct sheet *consOpen(int *fat)
{
	struct memory *mem = (struct memory *) MEMORY_ADDR;
	struct sheet *sht = shtAlloc();
	unsigned char *buf = (unsigned char *) memAlloc(mem, 256 * 165);
	struct task *task;
	shtSet(sht, buf, 256, 165, -1);
	winAlloc(sht, "R\[", 0);
	textbox(sht, _000000, 8, 28, 240, 128);
	task = consOpen0(sht, fat);
	sht->task = task;
	sht->flag |= SHEET_CUR;
	return sht;
}

struct task *consOpen0(struct sheet *sht, int *fat)
{
	struct memory *mem = (struct memory *) MEMORY_ADDR;
	struct task *task = taskAlloc();
	int *fifo = (int *) memAlloc(mem, 512);
	task->stack = memAlloc(mem, 65536);
	task->tss.esp = task->stack + 65536 - 8;
	task->tss.eip = (int) &consTask;
	task->tss.es = 1 * 8;
	task->tss.cs = 2 * 8;
	task->tss.ss = 1 * 8;
	task->tss.ds = 1 * 8;
	task->tss.fs = 1 * 8;
	task->tss.gs = 1 * 8;
	*((int *) (task->tss.esp + 4)) = (int) sht;
	fifoInit(&task->fifo, fifo, 128, task);
	task->fat = fat;
	taskRun(task, 2, 2);
	return task;
}

void consClose(struct sheet *sht)
{
	struct memory *mem = (struct memory *) MEMORY_ADDR;
	struct task *task = sht->task;
	memFree(mem, (int) sht->buf, 42240);
	shtFree(sht);
	consClose0(task);
	return;
}

void consClose0(struct task *task)
{
	extern struct tasks *tctl;
	struct memory *mem = (struct memory *) MEMORY_ADDR;
	taskSleep(task);
	memFree(mem, task->stack, 65536);
	memFree(mem, (int) task->fifo.buf, 512);
	cli();
	task->flag = 0;
	if (tctl->tfpu == task) {
		tctl->tfpu = 0;
	}
	sti();
	return;
}

void consPutc(struct console *cons, unsigned char c, char m)
{
	struct sheet *sht = cons->sht;
	unsigned char s[2];

	s[0] = c;
	s[1] = 0;
	if (s[0] == 0x09) {
		for (;;) {
			if (cons->sht != 0) {
				graphicBox(sht, _000000, cons->cx, cons->cy, cons->cx + 7, cons->cy + 15);
				graphicPuts(sht, _FFFFFF, cons->cx, cons->cy, " ");
				shtRefresh(sht, cons->cx, cons->cy, cons->cx + 8, cons->cy + 16);
			}
			cons->cx += 8;
			if (cons->cx == 248) {
				consNline(cons);
			}
			if (((cons->cx - 8) & 0x1f) == 0) {
				break;
			}
		}
	} else if (s[0] == 0x0a) {
		consNline(cons);
	} else if (s[0] == 0x0d) {
	} else {
		if (cons->sht != 0) {
			graphicBox(sht, _000000, cons->cx, cons->cy, cons->cx + 7, cons->cy + 15);
			graphicPuts(sht, _FFFFFF, cons->cx, cons->cy, s);
			shtRefresh(sht, cons->cx - 8, cons->cy, cons->cx + 8, cons->cy + 16);
		}
		if (m != 0) {
			cons->cx += 8;
			if (cons->cx == 248) {
				consNline(cons);
			}
		}
	}
	return;
}

void consPuts(struct console *cons, unsigned char *s, int n)
{
	int i;

	if (n != -1) {
		for (i = 0; i < n; i++) {
			consPutc(cons, s[i], 1);
		}
	} else {
		for (; *s != 0; s++) {
			consPutc(cons, *s, 1);
		}
	}
	return;
}

void consNline(struct console *cons)
{
	struct task *task = taskNow();
	struct sheet *sht = cons->sht;
	int x, y;
	if (cons->cy < 140) {
		cons->cy += 16;
	} else if (sht != 0) {
		for (y = 28; y < 140; y++) {
			for (x = 8; x < 248; x++) {
				sht->buf[y * sht->sx + x] = sht->buf[(y + 16) * sht->sx + x];
			}
		}
		for (y = 140; y < 156; y++) {
			for (x = 8; x < 248; x++) {
				sht->buf[y * sht->sx + x] = _000000;
			}
		}
		shtRefresh(sht, 8, 28, 248, 156);
	}
	cons->cx = 8;
	if (task->lbyte != 0) {
		cons->cx += 8;
	}
	return;
}

void consRuncmd(struct console *cons, int *fat)
{
	if (strcmp(cons->line, "mem") == 0 && cons->sht != 0) {
		consMem(cons);
	} else if (strcmp(cons->line, "dir") == 0 && cons->sht != 0) {
		consDir(cons);
	} else if (strcmp(cons->line, "exit") == 0) {
		consExit(cons, fat);
	} else if (strncmp(cons->line, "ncst ", 5) == 0) {
		consNcst(cons->line + 5, fat);
	} else if (strncmp(cons->line, "create ", 7) == 0) {
		consCreate(cons, fat);
	} else if (strncmp(cons->line, "delete ", 7) == 0) {
		consDelete(cons, fat);
	} else if (strncmp(cons->line, "rename ", 7) == 0) {
		consRename(cons);
	} else if (cons->line[0] != 0) {
		if (consApp(cons, fat) == 0) {
			consPuts(cons, "R}hsł܂B", -1);
		}
	}
	return;
}

void consMem(struct console *cons)
{
	struct memory *mem = (struct memory *) MEMORY_ADDR;
	unsigned char s[20];
	sprintf(s, " %4d/%4d MB", memStat(mem) / (1024 * 1024), mem->allsize / (1024 * 1024));
	consPuts(cons, s, -1);
	return;
}

void consDir(struct console *cons)
{
	struct file *file = (struct file *) (DISK_ADDR + 0x00002600);
	int i, j;
	unsigned char s[21];
	for (i = 0; i < 224; i++) {
		if (file[i].name[0] == 0x00) {
			break;
		}
		if (file[i].name[0] != 0xe5) {
			if ((file[i].type & 0x18) == 0) {
				sprintf(s, "FILENAME.EXT %7d\n", file[i].size);
				for (j = 0; j < 8; j++) {
					s[j] = file[i].name[j];
				}
				s[9] = file[i].ext[0];
				s[10] = file[i].ext[1];
				s[11] = file[i].ext[2];
				if (file[i].ext[0] == ' ') {
					s[8] = ' ';
				}
				consPuts(cons, s, -1);
			}
		}
	}
	return;
}

void consExit(struct console *cons, int *fat)
{
	struct task *task = taskNow();
	cli();
	if (cons->sht != 0) {
		consExitreq(cons->sht);
	} else {
		consExitreq0(task);
	}
	sti();
	for (;;) {
		taskSleep(task);
	}
}

void consNcst(unsigned char *line, int *fat)
{
	struct task *task = consOpen0(0, fat);
	struct fifo *fifo = &task->fifo;
	int i;
	for (i = 0; line[i] != 0; i++) {
		fifoPut(fifo, 256 + line[i]);
	}
	fifoPut(fifo, 256 + 10);
	return;
}

void consCreate(struct console *cons, int *fat)
{
	struct file *file = fileCreate(fat, cons->line + 7);
	if (file == 0) {
		consPuts(cons, "t@C쐬ł܂B", -1);
	}
	return;
}

void consDelete(struct console *cons, int *fat)
{
	struct file *file = fileSearch(cons->line + 7);
	if (file != 0) {
		fileDelete(fat, file);
	} else {
		consPuts(cons, "t@C܂B", -1);
	}
	return;
}

void consRename(struct console *cons)
{
	struct file *file;
	unsigned char *s;
	int i;
	for (s = cons->line + 7; *s != ' '; s++) { }
	*s = 0x00;
	s++;
	for (; *s == ' '; s++) { }
	file = fileSearch(cons->line + 7);

	if (file != 0) {
		i = fileRename(file, s);
		if (i == 0) {
			consPuts(cons, "t@Cł܂B", -1);
		}
	} else {
		consPuts(cons, "t@C܂B", -1);
	}
	return;
}

int consApp(struct console *cons, int *fat)
{
	struct file *file;
	struct task *task = taskNow();
	struct memory *mem = (struct memory *) MEMORY_ADDR;
	unsigned char s[13], *p, *q;
	int apps, segs, esp, dats, dhrb, ahrb, i;
	for (i = 0; i < 13; i++) {
		if (cons->line[i] <= ' ') {
			break;
		}
		s[i] = cons->line[i];
	}
	s[i] = 0;
	file = fileSearch(s);
	if (file == 0 && s[i - 1] != '.') {
		s[i] = '.';
		s[i + 1] = 'H';
		s[i + 2] = 'R';
		s[i + 3] = 'B';
		s[i + 4] = 0;
		file = fileSearch(s);
	}

	if (file != 0) {
		apps = file->size;
		p = fileRead(fat, file->cno, apps);
		if (apps >= 36 && strncmp(p + 4, "Hari", 4) == 0 && *p == 0x00) {
			segs = *((int *) p);
			esp = *((int *) (p + 0x0c));
			dats = *((int *) (p + 0x10));
			dhrb = *((int *) (p + 0x14));
			q = (unsigned char *) memAlloc(mem, segs);
			task->ds = (int) q;
			segdescSet(task->ldt, (int) p, file->size - 1, RXCODE + 0x60);
			segdescSet(task->ldt + 1, (int) q, segs - 1, RWDATA + 0x60);
			for (i = 0; i < dats; i++) {
				q[esp + i] = p[dhrb + i];
			}
			task->mem = 0;

			/* HariMainĂяoA߂ƂɏIpAPIĂяo */
			ahrb = *((int *) (p + 0x1c)) + 0x20 - 9;
			p[4] = 0xe8;
			p[5] = ahrb & 0xff;
			p[6] = (ahrb >> 8) & 0xff;
			p[7] = (ahrb >> 16) & 0xff;
			p[8] = (ahrb >> 24) & 0xff;
			p[9] = 0xb8;
			p[10] = 0x00;
			p[11] = 0x00;
			p[12] = 0x00;
			p[13] = 0x00;
			p[14] = 0xcd;
			p[15] = 0x30;
			appRun(0x04, 0 * 8 + 4, 1 * 8 + 4, esp, &(task->tss.esp0));

			shtClean(task);
			timerAllcancel(&task->fifo);
			beep(0);
			for (i = 0; i < 8; i++) {
				if (task->fh[i].buf != 0) {
					bufFlush(&task->fh[i]);
					memFree(mem, (int) task->fh[i].buf, 512);
					task->fh[i].buf = 0;
				}
			}
			task->lbyte = 0;
			memFree(mem, (int) q, segs);
		} else {
			consPuts(cons, "Ȏst@Cł͂܂B", -1);
		}
		memFree(mem, (int) p, apps);
		return 1;
	}
	return 0;
}

void consExitreq(struct sheet *sht)
{
	extern struct sheets *shts;
	struct fifo *fifo = (struct fifo *) *((int *) FIFO_ADDR);
	fifoPut(fifo, 768 + (sht - shts->sheet));
	return;
}

void consExitreq0(struct task *task)
{
	extern struct tasks *tctl;
	struct fifo *fifo = (struct fifo *) *((int *) FIFO_ADDR);
	fifoPut(fifo, 1024 + (task - tctl->task));
	return;
}

void consExitreq1(struct sheet *sht)
{
	extern struct sheets *shts;
	struct fifo *fifo = (struct fifo *) *((int *) FIFO_ADDR);
	fifoPut(fifo, 2024 + (sht - shts->sheet));
	return;
}
